
investments_list_default_columns = [
            "secid",
            "masterportfolioid",
            "tradingsymbol",
            "name",
            "securitytype",
            "exchangeid",
            "category",
        ]

# these datapointIds correspond to the above columns
investments_list_data_points = [
                {"datapointId": "OS00I", "isTsdp": False},
                {"datapointId": "DC09A", "isTsdp": False},
                {"datapointId": "OS385", "isTsdp": False},
                {"datapointId": "OS01W", "isTsdp": False},
                {"datapointId": "OS010", "isTsdp": False},
                {"datapointId": "OS01Z", "isTsdp": False},
                {"datapointId": "OS245", "isTsdp": False},
            ]


# Add custom error messages related to DataLake
UNAUTHORIZED_DATALAKE_REQUEST_ERROR = (
    "A valid license is required to query Morningstar datalake. Please fill out this form for further assistance: https://go.morningstar.com/LP=6808"
)
UNAVAILABLE_EXTERNALLY = "Morningstar datalake is only accessible through Analytics Lab"
DATA_NOT_FOUND = "The requested data could not be found. Check the query parameters and try again."
TEMP_TABLE_NAME_NOT_FOUND = "One or more temporary tables does not have a name. Check the query parameters and try again."



import pandas as pd
from pandas import DataFrame
import numpy as np
import warnings
from typing import List, Optional, Union, List, Dict, Any

from . import _decorator
from ..direct import investment
from ._exceptions import BadRequestException, ResourceNotFoundError
from . import _error_messages
from .data_type import Frequency, TimeSeriesFormat, CalculationWindowType
from ._config_key import FORMAT_DATE


@_decorator.typechecked
def returns(
    investments: Union[List[str], str, Dict[str, Any]],
    start_date: str = "2020-01-01",
    end_date: Optional[str] = None,
    freq: Union[Frequency, str] = Frequency.monthly,
    currency: Optional[str] = None,
) -> pd.DataFrame:
    warnings.warn(
        "The returns function is deprecated and will be removed in the next major version. Use get_returns instead",
        FutureWarning,
        stacklevel=2,
    )
    return get_returns(investments, start_date, end_date, freq, currency)


@_decorator.typechecked
def get_returns(
    investments: Union[List[str], str, Dict[str, Any]],
    start_date: str = "2020-01-01",
    end_date: Optional[str] = None,
    freq: Union[Frequency, str] = Frequency.monthly,
    currency: Optional[str] = None,
) -> pd.DataFrame:
    """A shortcut function to fetch return data for the specified investments.

    Args:
        investments (:obj:`Union`, `required`): Defines the investments to fetch. Input can be:

            * Investment IDs (:obj:`list`, `optional`): Investment identifiers, in the format of SecId;Universe or just SecId. E.g., ["F00000YOOK;FO","FOUSA00CFV;FO"] or ["F00000YOOK","FOUSA00CFV"]. Use the `investments <./lookup.html#morningstar_data.direct.investments>`_ function to discover identifiers.
            * Investment List ID (:obj:`str`, `optional`): Saved investment list in Morningstar Direct. Use the `get_investment_lists <./lists.html#morningstar_data.direct.user_items.get_investment_lists>`_ function to discover saved lists.
            * Search Criteria  ID (:obj:`str`, `optional`): Saved search criteria in Morningstar Direct. Use the `get_search_criteria <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria>`_ function to discover saved search criteria.
            * Search Criteria Condition (:obj:`dict`, `optional`): Search criteria definition. See details in the Reference section of `get_investment_data <#morningstar_data.direct.get_investment_data>`_ or use the `get_search_criteria_conditions <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria_conditions>`_ function to discover the definition of a saved search criteria.

        start_date (:obj:`str`): Start date of a date range for retrieving data. The format is
            YYYY-MM-DD, e.g., "2020-01-01".
        end_date (:obj:`str`, `optional`): End date of a date range for retrieving data. If no value is provided for
            end_date, current date will be used. The format is YYYY-MM-DD, e.g., "2020-01-01".
        freq (:obj:`md.direct.data_type.Frequency`): Enumeration of type md.direct.data_type.Frequency, which can be 'daily', 'weekly', 'monthly', 'quarterly', or 'yearly'. E.g., ``md.direct.data_type.Frequency.monthly``
        currency (:obj:`str`, `optional`): Three character code for the desired currency of returns, e.g., "USD".  Use the `currency_codes <./lookup.html#morningstar_data.lookup.currency_codes>`_ function to discover possible values.

    Returns:
        DataFrame: A DataFrame object with returns data.

    Examples:
        Get monthly returns.

    ::

        import morningstar_data as md


        df = md.direct.get_returns(
            investments=["F00000VKPI", "F000014B1Y"], start_date="2020-10-01", freq=md.direct.data_type.Frequency.monthly
        )
        df

    :Output:
        ==============  ==========  ====================================
        ID              Date          Monthly Return
        ==============  ==========  ====================================
        F00000VKPI      2020-10-31     -2.121865
        F00000VKPI      2020-11-30     6.337255
        F00000VKPI      2020-12-31     1.464777
        ...                ...         ...
        ==============  ==========  ====================================

    Errors:
        AccessDeniedError: Raised when the user is not properly authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user lacks permission to access a resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """

    if not isinstance(freq, Frequency):
        warnings.warn(
            "The use of string values for the 'freq' parameter is deprecated and will be removed in the next major version. Use Frequency enum values instead",
            FutureWarning,
            stacklevel=2,
        )
        freq = Frequency[freq]
    return_data_point = {"datapointId": freq.data_point_id, "isTsdp": True,
                         "startDate": pd.to_datetime(start_date).strftime(FORMAT_DATE)}
    if end_date:
        return_data_point["endDate"] = pd.to_datetime(end_date).strftime(FORMAT_DATE)
    if currency:
        return_data_point["currency"] = currency
    return investment.get_investment_data(investments=investments,
                                          data_points=[{"datapointId": "OS01W"}, return_data_point],
                                          time_series_format=TimeSeriesFormat.LONG_WITHOUT_NAME)


@_decorator.typechecked
def excess_returns(
    investments: Union[List, str, Dict[str, Any]],
    benchmark_sec_id: str,
    start_date: str = "2020-01-01",
    end_date: Optional[str] = None,
    freq: Union[Frequency, str] = Frequency.monthly,
    currency: Optional[str] = None,
) -> pd.DataFrame:
    warnings.warn(
        "The excess_returns function is deprecated and will be removed in the next major version. Use get_excess_returns instead",
        FutureWarning,
        stacklevel=2,
    )
    return get_excess_returns(investments, benchmark_sec_id, start_date, end_date, freq, currency)


@_decorator.typechecked
def get_excess_returns(
    investments: Union[List, str, Dict[str, Any]],
    benchmark_sec_id: str,
    start_date: str = "2020-01-01",
    end_date: Optional[str] = None,
    freq: Union[Frequency, str] = Frequency.monthly,
    currency: Optional[str] = None,
) -> pd.DataFrame:
    """A shortcut function to fetch excess return data for the specified investments.

    Args:
        investments (:obj:`Union`, `required`): Defines the investments to fetch. Input can be:

            * Investment IDs (:obj:`list`, `optional`): Investment identifiers, in the format of SecId;Universe or just SecId. E.g., ["F00000YOOK;FO","FOUSA00CFV;FO"] or ["F00000YOOK","FOUSA00CFV"]. Use the `investments <./lookup.html#morningstar_data.direct.investments>`_ function to discover identifiers.
            * Investment List ID (:obj:`str`, `optional`): Saved investment list in Morningstar Direct. Use the `get_investment_lists <./lists.html#morningstar_data.direct.user_items.get_investment_lists>`_ function to discover saved lists.
            * Search Criteria  ID (:obj:`str`, `optional`): Saved search criteria in Morningstar Direct. Use the `get_search_criteria <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria>`_ function to discover saved search criteria.
            * Search Criteria Condition (:obj:`dict`, `optional`): Search criteria definition. See details in the Reference section of `get_investment_data <#morningstar_data.direct.get_investment_data>`_ or use the `get_search_criteria_conditions <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria_conditions>`_ function to discover the definition of a saved search criteria.

        benchmark_sec_id (:obj:`str`): SecId of the security to use as the benchmark. Use the `investments <./lookup.html#morningstar_data.direct.investments>`_ function to discover identifiers.
        start_date (:obj:`str`): Start date of a date range for retrieving data. The format is
            YYYY-MM-DD, e.g., "2020-01-01".
        end_date (:obj:`str`, `optional`): End date of a date range for retrieving data. If no value is provided for
            end_date, current date will be used. The format is YYYY-MM-DD, e.g., "2020-01-01".
        freq (:obj:`md.direct.data_type.Frequency`): Enumeration of type md.direct.data_type.Frequency, which can be 'daily', 'weekly', 'monthly', 'quarterly', or 'yearly'. E.g., ``md.direct.data_type.Frequency.monthly``
        currency (:obj:`str`, `optional`): Three character code for the desired currency of returns, e.g., "USD".  Use the `currency_codes <./lookup.html#morningstar_data.lookup.currency_codes>`_ function to discover possible values.


    Returns:
        DataFrame: A DataFrame object with excess return data.

    Examples:
        Get monthly excess returns.

    ::

        import morningstar_data as md

        df = md.direct.get_excess_returns(
            investments=["F00000VKPI", "F000014B1Y"],
            benchmark_sec_id="F00000PLYW",
            freq=md.direct.data_type.Frequency.daily
        )
        df

    :Output:
        ==============  ==========  ====================================
        ID              Date          Monthly Return
        ==============  ==========  ====================================
        F00000VKPI      2020-10-31     -2.121865
        F00000VKPI      2020-11-30     6.337255
        F00000VKPI      2020-12-31     1.464777
        ...                ...         ...
        ==============  ==========  ====================================

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    if not isinstance(freq, Frequency):
        warnings.warn(
            "The use of string values for the 'freq' parameter is deprecated and will be removed in the next major version. Use Frequency enum values instead",
            FutureWarning,
            stacklevel=2,
        )
        freq = Frequency[freq]
        # This is a Direct custom calculation data point and measure of an investment's return in excess of a benchmark.
    return_data_point = {
        "datapointId": "20",
        "sourceId": freq.data_point_id,
        "windowType": CalculationWindowType.ROLLING_WINDOW.value,
        "annualized": False,
        "startDate": pd.to_datetime(start_date).strftime(FORMAT_DATE),
        "frequency": freq.abbr,
        "windowSize": 1,
        "stepSize": 1,
        "benchmark": benchmark_sec_id,
    }

    if currency:
        return_data_point["currency"] = currency
    if end_date:
        end_date = pd.to_datetime(end_date).strftime(FORMAT_DATE)
        return_data_point["endDate"] = end_date
    return investment.get_investment_data(investments=investments,
                                          data_points=[{"datapointId": "OS01W"}, return_data_point],
                                          time_series_format=TimeSeriesFormat.LONG_WITHOUT_NAME)



from enum import Enum
from typing import Any


class PortfolioType(Enum):
    def __init__(self, key: str, abbr: str) -> None:
        self.key = key
        self.abbr = abbr

    @classmethod
    def get_full_name_by_abbr(cls, abbr: str) -> str:
        # Api return values are "MD","AC","BM","UA", we use abbr to get the full portfolio type name.
        for member in cls._member_map_.values():
            if member._value_[1] == abbr:
                return member.name
        return ""

    # The key is used to create and get a portfolio.
    # The abbr is the portfolio list response,we can use this to get the full portfolio type name.
    model_portfolios = ("60", "MD")
    custom_benchmarks = ("63", "BM")
    client_accounts = ("90", "UA")



import logging
import os
from typing import Optional, List
from pandas import DataFrame
import json
from ..direct._config import _Config
from ..direct._backend_apis._signed_url_backend import SignedUrlBackend
import io

import pandas as pd
from enum import Enum
from ._data_objects import TempTable, CSVFile
from ._exceptions import TempTableNameNotFoundException

_logger = logging.getLogger(__name__)

DL_ACCOUNT = os.getenv("DL_ACCOUNT")

region = os.getenv("DL_AWS_REGION", "us-east-1")
is_external_user = os.getenv("IS_EXTERNAL_USER")
uim_user_id = os.getenv("UIM_USER_ID")
bucket_name = os.getenv("DL_TEMP_TABLE_BUCKET_NAME")
redshift_cluster_id = os.getenv("REDSHIFT_CLUSTER_ID")
redshift_database_name = os.getenv("REDSHIFT_DATABASE_NAME")


class Source(Enum):
    DATALAKE = 1
    LAKEHOUSE = 2


def query(query_str: str, temp_tables: Optional[List[TempTable]] = None) -> DataFrame:
    """
    :bdg-ref-danger:`Upcoming Feature <../upcoming_feature>`

    Retrieve the results of a SQL query from the Morningstar Data Lake.

    Args:
        query_str: SQL query to be executed in Morningstar Data Lake
        temp_tables: A list of temporary tables that will exist for the duration of the query.

    Returns:
        A DataFrame object with results of the SQL query.

    :Examples:

        Submit a query using a temp table

    ::

        import morningstar_data as md
        import pandas as pd


        df_my_table = pd.DataFrame({'sec_id': ['F0GBR0606A', 'F00000SYAH', 'F00000WP51'], 'closing_price': [128.372, 23.02, 528.33]})
        df_query = md.datalake.query(query_str = 'select * from my_table;', temp_tables = [md.datalake.TempTable('my_table', df_my_table)])

    :Output:

        ===========  ===============
         sec_id       closing_price
        ===========  ===============
        F0GBR0606A      128.372
        F00000SYAH       23.02
        F00000WP51      528.33
        ===========  ===============

    :Errors:
        InvalidQueryException: When ``query_str`` contains invalid SQL syntax.

        UnauthorizedDataLakeAccessError: When the calling user is not authorized to query the Morningstar Data Lake.

        TempTableNameNotFoundException: When one or more temp tables being used are not found in the query string.

    """
    signed_url_backend = SignedUrlBackend()
    config = _Config()

    if temp_tables:
        _logger.debug(f"Query function invoked with temp tables {temp_tables}")

        missing_tables = []
        for table in temp_tables:
            if table.name is None or table.name.strip() == "":
                _logger.debug("Temp table name is empty")
                raise TempTableNameNotFoundException from None

            if table.name not in query_str:
                missing_tables.append(table.name)

            body = {"is_temp": True, "table_name": table.name}

            # get a signed url to upload this table to
            _logger.debug(f"Uploading to stage table_name: {table.name}")
            resp = json.loads(signed_url_backend.do_post_request(url=config.temp_table_url(), data=body))

            table.table_stage_id = resp["table_id"]
            upload_url = resp["upload_url"]

            resp = signed_url_backend.put_csv_file(csv=CSVFile(table.df), signed_upload_url=upload_url)

            _logger.info(f"Upload complete under the following s3 id {table.table_stage_id} with url {upload_url}")

        if len(missing_tables) > 0:
            _logger.warn(f" Not all temp_tables are in query_str, make sure your query is correct. Missing table names: {missing_tables}")

    df = _get_query_as_df(signed_url_backend, config, query_str, temp_tables)

    return df


def _construct_query_endpoint_payload(query_str: str, temp_tables: Optional[List[TempTable]] = None) -> str:
    # Sample body from md package to POST /query endpoint

    # {
    #     "query": "select * from xyz join temp_a join temp_b;",
    #     "tables": [
    #         {"table_id": "sd98f-dsf89sdf-sdf-df", "table_name": "temp_a", "columns": [{"name": "col1", "type": "int64"}, {"name": "col2", "type": "float"}]},
    #         {"table_id": "ad98f-dsf89sdf-sdf-df", "table_name": "temp_b",
    # "columns": [{"name": "col1", "type": "int64"}, {"name": "col2", "type": "float"}]},
    #     ],
    # }

    # Within a Temp table the DF argument is not optional
    if temp_tables:
        payload = {
            "query": query_str,
            "tables": [
                {"table_id": table.table_stage_id, "table_name": table.name, "columns": [{"name": col, "type": str(table.df[col].dtype)} for col in table.df.columns]}  # type: ignore
                for table in temp_tables
            ],
        }
    else:
        payload = {"query": query_str}

    return json.dumps(payload)


def _get_query_as_df(signed_url_backend: SignedUrlBackend, config: _Config, query_str: str, temp_tables: Optional[List[TempTable]]) -> pd.DataFrame:

    payload = _construct_query_endpoint_payload(query_str, temp_tables)

    _logger.debug(f"Hitting the data lake query endpoint with this payload\n{payload}\n")

    download_url_response = signed_url_backend.do_post_request(url=config.query_api_url(), data=payload)
    download_url = json.loads(download_url_response)["url"]

    _logger.info(f"Downloading from s3 at the url {download_url}")

    resp = signed_url_backend.get_bytes(url=download_url)
    try:
        df = pd.read_csv(io.StringIO(resp.text))
    except pd.errors.EmptyDataError:
        _logger.warning(f"Empty Data Error for query {query}, returning an empty dataframe")
        return pd.DataFrame()

    return df



import re
import warnings
from typing import Any, Dict, Optional, Union
import requests

from pandas import DataFrame

from ..._base import _logger
from .. import _decorator, _error_messages, _utils
from .._base_api import APIBackend
from .._config import _Config
from .._exceptions import BadRequestException, ResourceNotFoundError, NetworkExceptionError

_config = _Config()


class DataSetAPIBackend(APIBackend):
    """
    Subclass to call the Data Set API and handle any HTTP errors that occur.
    """

    def __init__(self) -> None:
        super().__init__()

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json()["message"]
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_DATA_SET) from None
        else:
            pass


_data_set_api = DataSetAPIBackend()


@_decorator.not_null
def _get_morningstar_data_set_data_points(data_set_id: str) -> DataFrame:
    response_json = _get_defaultview_data_points(data_set_id)
    if not response_json.get("content", None):
        raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_DATA_SET)
    content = response_json["content"]
    extracted_content = _utils._extract_data(content)
    details = DataFrame(extracted_content)
    details_with_renamed_column = _utils._rename_data_frame_column(details, source="id", target="datapointId")
    return _utils._filter_data_frame_column_by_setting(details_with_renamed_column)


@_decorator.typechecked
def get_data_sets() -> DataFrame:
    """Returns all data sets saved by or shared to a user in Morningstar Direct.

    Returns:
        DataFrame: A DataFrame object with all data sets. DataFrame columns include:

        * datasetId
        * name
        * source
        * shared


    :Examples:

    ::

        import morningstar_data as md

        df = md.direct.user_items.get_data_sets()
        df

    :Output:
        ============  =====================  ============  ========
        datasetId     name                   source        shared
        ============  =====================  ============  ========
        5447361       alpha                  DESKTOP       False
        5386429       Michael's Search Ds	 DESKTOP       False
        ...
        ============  =====================  ============  ========

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    url = f"{_config.dataset_service_url()}v1/datasets"
    response_json: Dict[str, Any] = _data_set_api.do_get_request(url)
    if not response_json.get("datasets", None):
        raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_NO_DATA_SETS)

    result = DataFrame(response_json["datasets"])
    return result[["datasetId", "name", "source", "shared"]]


@_decorator.not_null
@_decorator.typechecked
def get_data_set(data_set_id: str) -> DataFrame:
    warnings.warn(
        "The get_data_set function is deprecated and will be removed in the next major version. Use get_data_set_details instead",
        FutureWarning,
        stacklevel=2,
    )
    return get_data_set_details(data_set_id)


@_decorator.not_null
@_decorator.typechecked
def get_data_set_details(data_set_id: str) -> DataFrame:
    """Returns all data points for a given saved data set.

    Args:
        data_set_id (:obj:`str`): Unique identifier of a Morningstar or user-created data set saved
            in Morningstar Direct, e.g., "6102286". Use the `get_data_sets <#morningstar_data.direct.user_items.get_data_sets>`_ or `get_morningstar_data_sets <#morningstar_data.direct.get_morningstar_data_sets>`_ functions to discover saved data sets.

    Returns:
        DataFrame: A DataFrame object with data points. DataFrame columns include:

        * datapointId
        * datapointName
        * displayName
        * currency
        * preEuroConversion
        * sourceId
        * frequency
        * startDate
        * endDate
        * floatStart
        * floatEnd
        * startDelay
        * endDelay
        * diffStart
        * diffEnd
        * compounding
        * calculationId
        * annualized
        * annualDays
        * benchmark
        * riskfree
        * windowType
        * windowSize
        * stepSize
        * requireContinueData
        * fit
        * scalType
        * scalValue
        * scalPercentValue
        * timehorizon


    Examples:
        Get data points contained in data set "0218-0450".

    ::

        import morningstar_data as md

        df = md.direct.user_items.get_data_set_details(data_set_id="0218-0450")
        df

    :Output:
        ===========  =====  ====  ========  =============  ===  ================================
        datapointId  alias  type  universe  datapointName  ...  calcIsApplyIsraelsenModification
        ===========  =====  ====  ========  =============  ===  ================================
        OS01W        Z0	    text            Name           ...  NaN
        ...
        ===========  =====  ====  ========  =============  ===  ================================

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    if re.match(r"^\d+$", data_set_id, re.M) is not None:
        return _get_user_data_set_data_points(data_set_id)
    elif re.match(r"^\d{4}-\d{4}$", data_set_id, re.M) is not None:
        return _get_morningstar_data_set_data_points(data_set_id)
    else:
        raise BadRequestException(_error_messages.BAD_REQUEST_ERROR_DATA_SET)


def _get_user_data_set_data_points(data_set_id: str) -> DataFrame:
    response_json = _get_user_data_set(data_set_id)
    if not response_json.get("content", None):
        raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_DATA_SET)
    content = response_json["content"]
    extracted_data = _utils._extract_data(content)
    details = DataFrame(extracted_data)
    return _utils._filter_data_frame_column_by_setting(details)


def _get_user_data_set(data_set_id: Optional[str]) -> Dict[str, Any]:
    url = f"{_config.dataset_service_url()}v1/datasets/{data_set_id}"
    response_json: Dict[str, Any] = _data_set_api.do_get_request(url)
    return response_json


def _get_defaultview_data_points(view_id: str) -> Dict[str, Any]:
    url = f"{_config.data_point_service_url()}v1/defaultviews/{view_id}/datapoints"
    response_json: Dict[str, Any] = _data_set_api.do_get_request(url)
    return response_json


def _get_binding_data_set(user_object_id: str, id_type: str) -> Dict[str, Any]:
    url = f"{_config.dataset_service_url()}v1/datasets/{id_type}/{user_object_id}"
    response_json: Dict[str, Any] = _data_set_api.do_get_request(url)
    return response_json



from .investment import investment_data, get_investment_data
from .returns import returns, get_returns, excess_returns, get_excess_returns
from .._base import _logger
from .holdings import holdings, get_holdings, holding_dates, get_holding_dates, get_lookthrough_holdings
from .performance_report import get_report, calculate_report, get_report_status, get_reports
from . import user_items
from .asset_flow import get_asset_flow, get_asset_flow_markets, get_asset_flow_data_points
from .lookup import get_morningstar_data_sets, investment_universes, get_data_point_settings, investments, firms, get_brandings, portfolio_managers, companies
from ._config import _Config
from .data_type import Frequency, Blank, PeerGroupMethodology, Order, TimeSeriesFormat
from ._portfolio_data_set import PortfolioDataSet
from .custom_database import my_database, firm_database
from . import portfolio
from .peer_group import get_peer_group_breakpoints

config = _Config()



from __future__ import annotations

import copy
from collections import OrderedDict 
from typing import Any, OrderedDict as OrderedDictType, List, NamedTuple
from dataclasses import dataclass, field
from pandas import DataFrame
import pandas as pd
from functools import reduce
from ..data_type import TimeSeriesFormat


class Column(NamedTuple):
    name: ColumnNameType
    value: ColumnValueType


InvestmentIdType = str
DataPointAliasType = str
ColumnNameType = Any
ColumnValueType = Any
ColumnListType = List[Column]
ColumnDictType = OrderedDictType[ColumnNameType, ColumnValueType]
MetaDataKeyType = Any
MetaDataValueType = Any


def pivot_time_series_data(df: DataFrame, id_vars: list = ["Id"], variable_column_name: str = "Date",
                           value_column_name: str = "value") -> DataFrame:
    """
    Pivot Time series dataframe that we receive from the DataFrame.
    """
    pivoted_df = pd.melt(df, id_vars=id_vars, var_name=variable_column_name, value_name=value_column_name)
    pivoted_df[variable_column_name] = pivoted_df[variable_column_name].str[-10:]
    pivoted_df[variable_column_name] = pd.to_datetime(pivoted_df[variable_column_name])
    pivoted_df = pivoted_df.sort_values(by=["Id", variable_column_name])

    return pivoted_df


@dataclass(frozen=True)
class InvestmentDataRequest:
    """Stores parameters for getting investment data from DO APIs"""

    investment_ids: list = field(default_factory=list)
    data_points: DataFrame = field(default_factory=DataFrame)


@dataclass
class InvestmentDataResults:
    """Stores multiple investment results returned by DO API, indexed by investment ID"""

    _data: OrderedDictType[InvestmentIdType, InvestmentDataResult] = field(default_factory=OrderedDict)

    # Commands
    def add_column_data(self, investment_id: InvestmentIdType, alias: DataPointAliasType, column_data: ColumnListType) -> None:
        self._ensure_investment_id_exists(investment_id)
        self._data[investment_id].add_column_data(alias, column_data)

    def add_meta_data(self, investment_id: InvestmentIdType, key: MetaDataKeyType, value: MetaDataValueType) -> None:
        self._ensure_investment_id_exists(investment_id)
        self._data[investment_id].add_meta_data(key, value)

    def merge_with(self, other_results: InvestmentDataResults, in_place: bool = False) -> InvestmentDataResults:
        new_results = self
        if not in_place:
            new_results = copy.deepcopy(self)

        for investment_id, other_result in other_results._data.items():
            new_results._ensure_investment_id_exists(investment_id)
            new_results._data[investment_id].merge_with(other_result)

        return new_results

    # Queries
    def get_column_data_value(self, investment_id: InvestmentIdType, alias: DataPointAliasType, col_name: ColumnNameType, default_value: ColumnValueType = None) -> ColumnValueType:
        return self._data.get(investment_id, InvestmentDataResult()).get_column_data_value(alias, col_name, default_value)

    def get_meta_data(self, investment_id: InvestmentIdType, key: MetaDataKeyType, default_value: MetaDataValueType = None) -> MetaDataValueType:
        return self._data.get(investment_id, InvestmentDataResult()).get_meta_data(key, default_value)

    def get_data_point_alias_to_col_dict(self) -> dict:
        if len(self._data) == 0:
            return dict()
        one_investment: InvestmentDataResult = list(self._data.values())[0]
        return one_investment.get_data_point_alias_to_col_dict()

    def as_list(self) -> list:
        new_list: list = []
        for investment_id, result in self._data.items():
            # TODO: find a better place for this check
            if result.get_meta_data("entitled", None) is False:
                continue
            new_item: dict = dict()
            # Always include the Investment ID column
            new_item["Id"] = investment_id
            for alias in result.get_aliases():
                new_item.update(result.get_column_data(alias))

            new_list.append(new_item)
        return new_list

    def as_long_frame(self) -> DataFrame:
        all_investments_data_frame = DataFrame()
        for investment_id, result in self._data.items():
            all_df = DataFrame()
            ts_df = DataFrame()
            non_ts_data: dict = dict()
            df_list = []
            data_point_cols = []
            # Always include the Investment ID column
            non_ts_data["Id"] = investment_id
            for alias in result.get_aliases():
                # get investment_id data point values dict
                data_point_value_dict = result.get_column_data(alias)

                # get investment_id data point meta
                data_point_meta_data = result.get_meta_data(alias, None)
                isTsdp = data_point_meta_data.get("isTsdp", None)
                data_point_name = data_point_meta_data.get("name", data_point_meta_data.get("datapointName", None))
                # columns order
                data_point_cols.append(data_point_name)
                if isTsdp:
                    # pivot ts data
                    data_point_value_dict["Id"] = investment_id
                    df_list.append(
                        pivot_time_series_data(DataFrame([data_point_value_dict]), value_column_name=data_point_name))
                    # merge ts data
                    ts_df = reduce(lambda left, right: pd.merge(left, right, on=["Id", "Date"], how="outer"), df_list)
                else:
                    non_ts_data.update(data_point_value_dict)
            # merge ts and non ts
            all_df = pd.merge(DataFrame([non_ts_data]), ts_df, on=["Id"],
                              how="outer") if not ts_df.empty else DataFrame([non_ts_data])
            # reorder columns
            all_df = all_df[["Id", "Date"] + data_point_cols] if not ts_df.empty else all_df[data_point_cols]
            all_investments_data_frame = pd.concat([all_investments_data_frame, all_df])

        return all_investments_data_frame.reset_index(drop=True)

    def as_data_frame(self, order_cols_by: list = [],
                      time_series_format: TimeSeriesFormat = TimeSeriesFormat.WIDE) -> DataFrame:
        if time_series_format == time_series_format.LONG:
            return self.as_long_frame()
        if time_series_format == time_series_format.LONG_WITHOUT_NAME:
            return self.as_long_frame().drop(columns=["Name"])

        new_df = DataFrame(self.as_list())
        if order_cols_by:
            # Always include the Investment ID column
            if "Id" not in order_cols_by: 
                order_cols_by=["Id"] + order_cols_by
            return new_df[order_cols_by]
        else:
            return new_df

    # Helpers
    def _ensure_investment_id_exists(self, investment_id: InvestmentIdType) -> None:
        if investment_id not in self._data:
            self._data[investment_id] = InvestmentDataResult()


@dataclass
class InvestmentDataResult:
    """Stores one investment result containing multiple data points, indexed by data point alias.
    Each data point can contain one or more values (e.g., if data point contains time-series data).
    Each value represents a column."""

    _data: OrderedDictType[DataPointAliasType, ColumnDictType] = field(default_factory=OrderedDict)
    _meta_data: OrderedDictType[MetaDataKeyType, MetaDataValueType] = field(default_factory=OrderedDict)

    # Commands
    def add_column_data(self, alias: DataPointAliasType, column_data: ColumnListType) -> None:
        self._ensure_alias_exists(alias)
        self._data[alias].update(column_data)

    def add_meta_data(self, key: MetaDataKeyType, value: MetaDataValueType) -> None:
        self._meta_data[key] = value

    def merge_with(self, other_result: InvestmentDataResult) -> None:
        for alias, other_values in other_result._data.items():
            self._ensure_alias_exists(alias)
            self._data[alias].update(other_values)
        self._meta_data.update(other_result._meta_data)

    # Queries
    def get_column_data(self, alias: DataPointAliasType, default_value: ColumnDictType = OrderedDict()) -> ColumnDictType:
        return self._data.get(alias, default_value)

    def get_column_data_value(self, alias: DataPointAliasType, col_name: ColumnNameType, default_value: ColumnValueType = None) -> ColumnValueType:
        return self.get_column_data(alias).get(col_name, default_value)

    def get_meta_data(self, key: MetaDataKeyType, default_value: MetaDataValueType = None) -> MetaDataValueType:
        return self._meta_data.get(key, default_value)

    def get_data_point_alias_to_col_dict(self) -> dict:
        result = dict()
        for alias, columns in self._data.items():
            result[alias] = list(columns.keys())
        return result

    def get_aliases(self) -> list:
        return list(self._data.keys())

    # Helpers
    def _ensure_alias_exists(self, alias: DataPointAliasType) -> None:
        if alias not in self._data:
            self._data[alias] = OrderedDict()



import simplejson as json
import logging
import os
from typing import Dict, Any, Optional
import polling2

from botocore.exceptions import ClientError
from ..direct._backend_apis import DeliveryAPIBackend

from ._delivery_config import DeliveryConfig
from ._helpers import _get_mds_base_url, _get_mds_bucket, _create_deliver_body, _create_feed_body, _create_output_file_name, _validate_file_name
from ..direct import _decorator

_logger = logging.getLogger(__name__)
_api_backend = DeliveryAPIBackend()


def _create_feed_id() -> int:
    """Calls the MDS feeds api to generate a new feed id.

    This feed id is needed when calling the MDS deliver api. In general, when a notebook is running
    in a schedule it will be passed in this feed id. But for an ad-hoc delivery this feed id needs
    to be created before the MDS delivery api can be called.
    """
    _logger.info("Creating a feed id using MDS api")
    url = f"{_get_mds_base_url()}v1/feeds?type=notebook"
    data = _create_feed_body()

    try:
        _logger.info(f"Calling MDS /feeds endpoint with: {data}")
        response_json = _api_backend.do_post_request(url, data=json.dumps(data, ignore_nan=True))
        feed_id: int = response_json["feedId"]
        return feed_id
    except Exception:
        _logger.error("Could not create a feed id using MDS api")
        raise


def _upload_to_s3(file_path: str, output_file_name: str) -> None:
    """Uploads a file to the MDS s3 bucket"""
    try:
        output_bucket = _get_mds_bucket()
        output_bucket.upload_file(file_path, output_file_name)
    except FileNotFoundError:
        _logger.error(f"File not found: {file_path}, make sure the file you are trying to upload exists")
        raise
    except ClientError:
        _logger.error("Could not upload to s3 bucket due to AWS error")
        raise
    except Exception:
        _logger.error("Could not upload to s3 bucket")
        raise


def _mds_deliver_file(output_file_name: str, config: DeliveryConfig, feed_id: int, delivered_file_name: Optional[str] = None) -> str:
    """Calls the MDS delivery api with the needed information to deliver the uploaded file and returns job_id"""
    url = f"{_get_mds_base_url()}v1/delivery/{feed_id}"
    data = _create_deliver_body(output_file_name, config, delivered_file_name)
    _logger.info(f"Calling MDS /delivery/{feed_id} endpoint with: {data}")
    response_json = _api_backend.do_post_request(url, data=json.dumps(data, ignore_nan=True))
    _logger.info(f"Delivery response: {response_json}")
    job_id: str = response_json["jobId"]
    return job_id


@_decorator.typechecked
def get_delivery_profile() -> Dict:
    """Fetch ftp delivery profiles using MDS API
    Returns:
        :obj:`dict`: FTP delivery profiles for the user.
    """
    url = f"{_get_mds_base_url()}v1/delivery-profile"
    _logger.info(f"Calling MDS /delivery-profile endpoint. ")
    mds_delivery_profiles: Dict[Any, Any] = _api_backend.do_get_request(url)
    _logger.info(f"MDS delivery profile response: {mds_delivery_profiles}")
    ftp_delivery_profiles = {"ftps": mds_delivery_profiles.get("ftps", [])}
    return ftp_delivery_profiles


@_decorator.typechecked
def delivery_status(job_id: str, poll: bool = True) -> str:
    """Polls the MDS delivery api with the job_id to check the delivery status.
    Args:
        job_id (:obj:`str`): The job id that we get from MDS endpoint
        poll (:obj:`bool`): Flag to poll the endpoint or not (default is set as True,
        as we want to poll the endpoint to get the status)

    Returns:
        :obj:`str`: Delivery status as given by the endpoint.
    """
    url = f"{_get_mds_base_url()}v1/jobs/{job_id}/status"
    _logger.info(f"Calling MDS /delivery/{job_id}/status endpoint")

    # polling the endpoint till it gets status response; raise exception error if occurred
    def _delivery_finished(res: dict) -> bool:
        try:
            return res["jobStatus"].lower() in ["done", "failed"]
        except KeyError as e:
            _logger.error(f"Could not find delivery status. Exception as {e}")
            raise

    if poll:
        _logger.info(f"Polling the request to get the delivery status.")
        try:
            polling2.poll(
                lambda: _api_backend.do_get_request(url),
                step=5,
                timeout=60,
                check_success=lambda response: _delivery_finished(response),
            )
        except polling2.TimeoutException as te:
            _logger.info(f"Timeout for the request. Exception as {te}.")

    response_json = _api_backend.do_get_request(url)
    _logger.info(f"Delivery status: {response_json}")
    return str(response_json["jobStatus"])  # "DONE" or "FAILED" or "RUNNING"


@_decorator.typechecked
def deliver(file_path: str, config: DeliveryConfig, wait_for_delivery: bool = False, delivered_file_name: Optional[str] = None) -> dict:
    """
    :bdg-ref-danger:`Upcoming Feature <../upcoming_feature>`

    Delivers a file from a notebook to an email or FTP.

    Args:
        file_path (:obj:`str`): The path to the file that will be delivered, including file name and extension.
        config (:obj:`md.utils.DeliveryConfig`): An object that holds the delivery configuration information.

            For an email it contains
                * method (:obj:`DeliveryType`): ``DeliveryType.EMAIL`` in this case.
                * email_address (:obj:`str`): email_address to send the file to.
            For FTP it contains
                * method (:obj:`DeliveryType`): ``DeliveryType.FTP`` in this case.
                * user_name (:obj:`str`): the user_name to login to the ftp server.
                * password (:obj:`str`): the password to login to the ftp server.
                * server (:obj:`str`): the ftp server to use.
                * folder (:obj:`str`): the folder on the ftp server to upload the file to.
            For a Delivery Profile it contains
                * method (:obj:`DeliveryType`): ``DeliveryType.DELIVERY_PROFILE`` in this case.
                * delivery_profile_id (:obj:`str`): delivery_profile_id that was setup by MDS. Can be retrieved with get_delivery_profile()
        wait_for_delivery (:obj:`bool`): Flag to poll the api to check the delivery status or not. (True=show status)
        delivered_file_name (:obj:`Optional[str]`): An optional parameter for the user to specify the file name to be used when
            delivered. No file extension should be provided. Valid characters are `., -, _, (, ), , a-z, A-Z, 0-9`. If the file
            name includes non-valid characters, an exception will be raised.

    Returns:
        :obj:`dict`: A dictionary with the 'job_id', 'message', 'delivery_status' keys containing information about the delivery status

    :Examples:

    Deliver to an email address.

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        # Email example
        delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address="test@email.com")
        md.utils.deliver("test_export.csv", delivery_config)

    Deliver to a ftp server.

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        # FTP example
        delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.FTP, user_name="test", password="test", server="ts.ftp.com", folder="data/")
        md.utils.deliver("test_export.csv", delivery_config)

    Deliver to a Delivery Profile

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        # Delivery Profile example
        delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.DELIVERY_PROFILE, delivery_profile_id="1234")
        md.utils.deliver("test_export.csv", delivery_config)

    Multiple delivery within a single notebook.

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        ftp_delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.FTP, user_name="test", password="test", server="ts.ftp.com", folder="data/")
        md.utils.deliver("test_export.csv", ftp_delivery_config)

        email_delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address="test@email.com")
        md.utils.deliver("test_export.csv", email_delivery_config)

    Deliver with a specific filename.

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        # Email example
        delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address="test@email.com")
        md.utils.deliver("test_export.csv", delivery_config, delivered_file_name="Delivered_Dataframe_123")
        # File delivered will be named "Delivered_Dataframe_123.csv"

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        UnauthorizedDeliveryException: Raised when the user does not have permissions to deliver artifacts.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        FileNotFoundError: Raised when the file path specified to be delivered does not exist
    """
    if delivered_file_name:
        _validate_file_name(delivered_file_name)

    # Check for a feed_id, create one if not specified
    try:
        feed_id = int(os.environ["FEED_ID"])
    except KeyError:
        feed_id = _create_feed_id()

    output_file_name = _create_output_file_name(file_path)

    _upload_to_s3(file_path, output_file_name)
    job_id: str = _mds_deliver_file(output_file_name, config, feed_id, delivered_file_name)

    res = {"job_id": job_id, "message": "Delivery has been submitted."}

    # if flag is set True, poll the api, return the delivery status
    res["delivery_status"] = delivery_status(job_id, poll=wait_for_delivery)
    return res



from ast import arg
from asyncio.log import logger
from json import load
import logging


logging.basicConfig(level=logging.WARNING)
logging.getLogger("requests").setLevel(logging.CRITICAL)
logging.getLogger("boto3").setLevel(logging.CRITICAL)
logging.getLogger("botocore").setLevel(logging.CRITICAL)
logging.getLogger("uvicorn").setLevel(logging.ERROR)
logging.getLogger("fastapi").setLevel(logging.CRITICAL)
logging.getLogger("urllib3.connectionpool").setLevel(logging.CRITICAL)
logging.getLogger("celery").setLevel(logging.CRITICAL)
_logger = logging.getLogger(__name__)
_logger.propagate = True



import requests
import shutil

from pandas import DataFrame
from typing import List, Optional, Dict, Any, Union, Callable

from datetime import datetime
from urllib.parse import quote

from .._base import _logger
from . import _decorator
from . import _error_messages
from ._config import _Config
from ._core_api import make_request
from ._exceptions import BadRequestException, NetworkExceptionError, AccessDeniedError, ValueErrorException, ResourceNotFoundError
from ._error_messages import RESOURCE_NOT_FOUND_ERROR_PERFORMANCE_REPORT
from ._base_api import APIBackend

_config = _Config()


class PerformanceReportAPIBackend(APIBackend):
    """
    Subclass to call the Performance Report API and handle any HTTP errors that occur.
    """

    def __init__(self) -> None:
        super().__init__()

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json().get("message", "")
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(RESOURCE_NOT_FOUND_ERROR_PERFORMANCE_REPORT) from None


_performance_report_api_request = PerformanceReportAPIBackend()


@_decorator.typechecked
def get_reports() -> DataFrame:
    """Returns all performance reports saved or shared to a user in Morningstar Direct.

    Returns:
        DataFrame: A DataFrame object with all performance reports. DataFrame columns include:

        * reportId
        * name
        * permission
        * ownerId
        * shared
        * createdOn
        * lastCalculatedOn
        * folderId

    :Examples:

    ::

        import morningstar_data as md

        df = md.direct.performance_report.get_reports()
        df

    :Output:
        ========  ======  ==========  =======  ======  ====================  ================  ========
        reportId  name    permission  ownerId  shared  createdOn             lastCalculatedOn  folderId
        ========  ======  ==========  =======  ======  ====================  ================  ========
        4940775   sample  READ_WRITE  XXX      False   2017-08-17T09:39:00Z                    1
        ...
        ========  ======  ==========  =======  ======  ====================  ================  ========

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

    """
    url = f"{_config.performancereport_service_url()}v1/reports"
    response_json: Dict[str, Any] = _performance_report_api_request.do_get_request(url)
    return DataFrame(response_json["reports"])


@_decorator.not_null
@_decorator.typechecked
def get_report(report_id: str) -> DataFrame:
    """Returns performance report data for the specified report ID.

    Args:
        report_id (:obj:`str`): Unique identifier of a saved performance report from the Performance Reporting
            module in Morningstar Direct, e.g., "7782164". Use `get_reports <#morningstar_data.direct.get_reports>`_ to discover possible values.

    Returns:
        DataFrame: A DataFrame object with calculated performance report data. Columns include all columns
        that the user configured in the performance report.

    Examples:
        Get performance report data.

    ::

        import morningstar_data as md

        df = md.direct.performance_report.get_report(report_id="2463866") # Replace with a valid report ID
        df

    :Output:
        ================  ===========  =============  =========================  ======
        Group/Investment  Object Type  Display Group  Peer Group                 Ticker
        ================  ===========  =============  =========================  ======
        sample            investments  Unclassified   Peer Group: Display Group  XXX
        ================  ===========  =============  =========================  ======

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.
    """
    url = f"{_config.performancereport_service_url()}v1/reports/{report_id}"
    response_json: Dict[str, Any] = _performance_report_api_request.do_get_request(url)
    return _merge_datas(response_json)


def _get_all_columns(columns: List[Dict[str, Any]], name_master: Optional[str] = None) -> List[Dict[str, Any]]:
    all_columns = []
    for column in columns:
        if "hidden" in column and column["hidden"] is True:
            continue
        elif "children" in column:
            all_columns += _get_all_columns(column["children"], column["name"])
        elif name_master is None and column["name"] == "Name":
            column["fullName"] = "Group/Investment"
            all_columns.append(column)
            all_columns.append({"fullName": "Object Type", "id": "ObjectType"})
            all_columns.append({"fullName": "Display Group", "id": "DisplayGroup"})
            all_columns.append({"fullName": "Peer Group", "id": "PeerGroup"})
        else:
            column["fullName"] = name_master + " " + column["name"] if name_master is not None else column["name"]
            if column.get("mstarIpType", "") == "":
                all_columns.append(column)
            else:
                all_columns.append({"fullName": column["fullName"], "id": column["id"]})
                all_columns.append(
                    {
                        "fullName": column["fullName"] + " - display text",
                        "id": column["id"],
                        "ismstarip": True,
                    }
                )

    return all_columns


def _get_all_columns_name(columns: List[Dict[str, Any]]) -> List[str]:
    all_columns = []
    for c in columns:
        all_columns.append(c["fullName"])
    return all_columns


def _filter_grouy_by_type(groups: List[Dict[str, Any]], type: str) -> Dict[str, Any]:
    get_by_type: Callable = lambda group: group["type"] == type
    get_id: Callable = lambda group: group["id"]
    new_groups = list(filter(get_by_type, groups))
    groups_group_id = list(map(get_id, new_groups))
    groups_group_dict = dict(zip(groups_group_id, new_groups))
    return groups_group_dict


def _get_data_dict(datas: Dict[str, Any], name: str) -> Dict[str, Any]:
    get_id: Callable = lambda datas: datas[name]
    datas_id = list(map(get_id, datas))
    return dict(zip(datas_id, datas))


def _set_data_rows(values: Dict[str, Any], columns: List[Any], switch: Dict[str, Any]) -> List[Any]:
    set_data = []
    for col in columns:
        col_id = col["id"]
        to_add = switch.get(col_id, None)
        if to_add is not None:
            set_data.append(to_add)
        elif col_id in values:
            if col.get("ismstarip", False):
                set_data.append(values[col_id].get("text", ""))
            else:
                set_data.append(values[col_id].get("value", ""))
        else:
            set_data.append("")

    return set_data


def _get_data_rows(
    investments: List[Dict[str, Any]],
    groups_group: Dict[str, Any],
    new_columns: List[Dict[str, Any]],
    ObjectType: Any,
    PeerGroup: Any,
) -> List[Any]:
    output_datas = []
    for inv in investments:
        group_id = inv["groupId"]
        values_dict = _get_data_dict(inv["values"], "alias")
        switch = {
            "ObjectType": ObjectType,
            "DisplayGroup": groups_group[group_id]["name"],
            "PeerGroup": PeerGroup,
        }
        output_datas.append(_set_data_rows(values_dict, new_columns, switch))
    return output_datas


def _filter_datas_by_group_id(datas: List[Dict[str, Any]], group_id: str) -> List[Dict[str, Any]]:
    get_by_id = lambda data: data["groupId"] == group_id
    new_datas = list(filter(get_by_id, datas))
    return new_datas


def _get_sub_group(groups_data: List[Dict[str, Any]], group_id: str) -> Dict[str, Any]:
    get_id = lambda group: group["id"]
    get_by_group_id = lambda data: "groupId" in data and data["groupId"] == group_id
    new_groups = list(filter(get_by_group_id, groups_data))
    groups_group_id = list(map(get_id, new_groups))
    return dict(zip(groups_group_id, new_groups))


def _get_group_data_rows(
    datas: List[Dict[str, Any]],
    groups_group: Any,
    new_columns: List[Dict[str, Any]],
    ObjectType: Any,
    group_name: Any,
) -> List[Any]:
    output_datas = []
    switch = {
        "ObjectType": ObjectType,
        "DisplayGroup": group_name,
        "PeerGroup": "",
    }

    for inv in datas:
        values_dict = _get_data_dict(inv["values"], "alias")
        output_datas.append(_set_data_rows(values_dict, new_columns, switch))

    return output_datas


def _get_peer_group_name_by_group_id(group_peer_groups: Union[Dict[Any, Any], Any]) -> Any:

    if group_peer_groups and group_peer_groups[0] and group_peer_groups[0]["values"] and group_peer_groups[0]["values"][0] and group_peer_groups[0]["values"][0]["value"]:
        return group_peer_groups[0]["values"][0]["value"]

    return ""


def _merge_datas(performance_report_data: Dict[str, Any]) -> DataFrame:
    metadata = performance_report_data["metaData"]
    view = metadata["view"]
    if view and view["id"] == "5":
        raise BadRequestException("This view not supported: " + view["name"])

    subtype_name = {
        "PeerGroup": "Peer Group Statistics",
        "DisplayGroup": "Display Group Statistics",
    }

    columns = performance_report_data["columns"]
    investments = performance_report_data["investments"]
    benchmarks = performance_report_data["benchmarks"]
    groups_data = performance_report_data["groups"]
    peer_groups = performance_report_data["peerGroups"]
    ranks = performance_report_data["ranks"]
    summary_statistics = performance_report_data["summaryStatistics"]

    all_data_row = []
    new_columns = _get_all_columns(columns)
    new_columns_name = _get_all_columns_name(new_columns)

    get_by_type = lambda group: group["type"] == "Group"
    new_groups = list(filter(get_by_type, groups_data))
    groups_group = _filter_grouy_by_type(groups_data, "Group")
    for group in new_groups:
        group_id = group["id"]
        group_name = group["name"]
        sub_groups = _get_sub_group(groups_data, group_id)
        group_investments = _filter_datas_by_group_id(investments, group_id)
        group_benchmarks = _filter_datas_by_group_id(benchmarks, group_id)
        group_peer_groups = _filter_datas_by_group_id(peer_groups, group_id)
        group_ranks = _filter_datas_by_group_id(ranks, group_id)
        peer_group_name = _get_peer_group_name_by_group_id(group_peer_groups)
        all_data_row.extend(
            _get_data_rows(
                group_investments,
                groups_group,
                new_columns,
                "investments",
                peer_group_name,
            )
        )
        all_data_row.extend(_get_data_rows(group_benchmarks, groups_group, new_columns, "Display Benchmark", ""))
        all_data_row.extend(_get_group_data_rows(group_peer_groups, groups_group, new_columns, "PeerGroup", group_name))
        all_data_row.extend(_get_group_data_rows(group_ranks, groups_group, new_columns, "Ranks", group_name))

        for key in sub_groups:
            subtype = sub_groups[key]["type"]

            if subtype in subtype_name:
                sub_peer_groups = _filter_datas_by_group_id(summary_statistics, sub_groups[key]["id"])
                all_data_row.extend(
                    _get_group_data_rows(
                        sub_peer_groups,
                        sub_peer_groups,
                        new_columns,
                        subtype_name[subtype],
                        group_name,
                    )
                )

    lssilss = _filter_datas_by_group_id(summary_statistics, "lssilss")
    if lssilss:
        all_data_row.extend(_get_group_data_rows(lssilss, lssilss, new_columns, "List Summary Statistics", ""))

    data_frame = DataFrame(all_data_row)
    data_frame.columns = new_columns_name
    return data_frame


@_decorator.not_null
@_decorator.typechecked
def calculate_report(report_id: str) -> DataFrame:
    """Initiates re-calculation of a performance report.

    Args:
        report_id (:obj:`str`): Unique identifier of a saved performance report from the Performance Reporting
            module in Morningstar Direct, e.g., "7782164". Use `get_reports <#morningstar_data.direct.get_reports>`_ to discover possible values.

    Returns:
        DataFrame: A DataFrame object confirming that report calculation was triggered. DataFrame columns include:

         * reportId
         * success

    Raises:
        ValueErrorException: Raised when the ``report_id`` parameter is invalid.

    Examples:
        Calculate the performance report.

    ::

        import morningstar_data as md

        df = md.direct.performance_report.calculate_report(report_id="7128568") # Replace with a valid report ID
        df

    :Output:
        ========  =======
        reportId  success
        ========  =======
        7128568   True
        ========  =======

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.
    """
    df = get_report_status(report_id)
    if df["status"].values[0] == "Ready" or df["status"].values[0] == "Failed":
        url = f"{_config.performancereport_service_url()}v1/reports/calc/{report_id}"
        response_json = _performance_report_api_request.do_post_request(url)
        return DataFrame(response_json)
    else:
        raise ValueErrorException(f"report {report_id} is calculating.")


@_decorator.not_null
@_decorator.typechecked
def get_report_status(report_id: str) -> DataFrame:
    """Returns the current calculation status of the specified performance report. Possible statuses are:

    * Failed
    * Queued
    * Calculating
    * Generating Report
    * Merging Excel
    * Downloading
    * Ready

    Args:
        report_id (:obj:`str`): Unique identifier of a saved performance report from the Performance Reporting
            module in Morningstar Direct, e.g., "7782164". Use `get_reports <#morningstar_data.direct.get_reports>`_ to discover possible values.

    Returns:
        DataFrame: A DataFrame object with calculation status. DataFrame columns include:

        * reportId
        * status

    Examples:
        Get performance report calculation status.

    ::

        import morningstar_data as md

        df = md.direct.performance_report.get_report_status(report_id="2463866") # Replace with a valid report ID
        df

    :Output:
        ========  ======
        reportId  status
        ========  ======
        2463866   Ready
        ========  ======

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.
    """
    url = f"{_config.performancereport_service_url()}v1/reports/status/{report_id}"
    response_json = _performance_report_api_request.do_get_request(url)
    return DataFrame(response_json)


@_decorator.not_null
@_decorator.typechecked
def export_to_excel(report_id: str, file_name: Optional[str] = None) -> str:
    """Export a performance report to Excel.

    Args:
        report_id (:obj:`str`): Unique identifier of a saved performance report from the Performance Reporting
            module in Morningstar Direct, e.g., "7782164". Use `get_reports <#morningstar_data.direct.get_reports>`_ to discover possible values.
        file_name (:obj:`str`, `optional`): Custom file name for the exported Excel file. If not provided, a default name with the format "report_<report_id>" will be used, e.g., "report_7782164.xlsx".

    Returns:
        str: Excel file name of the exported performance report.


    Examples:
        Export the given performance report.

    ::

        import morningstar_data as md

        resp = md.direct.performance_report.export_to_excel(report_id="2442678") # Replace with a valid report ID
        resp

    :Output:
        report_2442678.xlsx


    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    report_id_parse = quote(report_id, "utf-8")
    url = f"{_config.performancereport_export_service_url()}v1/reports/export/{report_id_parse}"
    presigned_url_response = _performance_report_api_request.do_get_request(url=url)

    current_date_time = datetime.now().strftime("%Y%m%d_%H%M%S")
    excel_file_suffix = ".xlsx"
    if file_name is None or file_name == "":
        file_name = f"performance_report_{report_id}_{current_date_time}.xlsx"
    elif not file_name.endswith(excel_file_suffix):
        file_name = file_name + excel_file_suffix
    try:
        url = presigned_url_response["url"]
        response = requests.get(url, stream=True)
        response.raise_for_status()
        with open(file_name, "wb") as excel_file:
            response.raw.decode_content = True
            shutil.copyfileobj(response.raw, excel_file)
        return file_name
    # Catching HTTPErrors as this is presigned URL from bucket no do_get_request is used
    except requests.ConnectionError as e:
        _logger.error(e)
        raise NetworkExceptionError from None
    except requests.HTTPError as e:
        _performance_report_api_request._handle_custom_http_errors(response)
        _performance_report_api_request._handle_default_http_errors(response)
        raise e



import re
from ._table import Table

from pandas import DataFrame
from typing import Any, Optional


class TempTable(Table):
    def __init__(self, name: str, df: DataFrame, table_stage_id: Optional[str] = None) -> None:
        super().__init__(name, df)
        self.table_stage_id = table_stage_id

    def _validate_column_names(self, df: DataFrame) -> DataFrame:
        columns = df.columns

        for col in columns:
            if len(col) > 127:
                raise ValueError(f"Invalid column name {col}: Column names must be less than 128 characters in length.")
            if not (col[0] == "_" or col[0].isalpha()):
                raise ValueError(f"Invalid column name {col}: Column names can only start with a letter or underscore.")
            if not re.search("^[\w|\$]*$", col):
                raise ValueError(f"Invalid column name {col}: Column names can only contain alphanumeric characters, underscores (_) or dollar signs ($).")

            return df

    def get_create_query(self) -> Any:
        query_string = f'CREATE TEMP TABLE "{self.name}" ('
        for col_type, col_name in zip(self.df.dtypes, self.df):  # type: ignore
            converted_type = self._pd_dtype_to_redshift_dtype(col_type.name)
            query_string += f'\n"{col_name}" {converted_type},'

        query_string = query_string[:-1] + "\n)"

        return query_string

    def get_insert_query(self, bucket_path: str, bucket_name: Optional[str], iam_role: Optional[str]) -> str:  # type: ignore
        query = f"""
                COPY {self.name}
                FROM 's3://{bucket_name}/{bucket_path}'
                IAM_ROLE '{iam_role}'
                CSV;
                """
        return query

    def _pd_dtype_to_redshift_dtype(self, dtype: Any) -> str:
        if dtype.startswith("int64"):
            return "BIGINT"
        elif dtype.startswith("int"):
            return "INTEGER"
        elif dtype.startswith("float"):
            return "REAL"
        elif dtype.startswith("datetime"):
            return "TIMESTAMP"
        elif dtype == "bool":
            return "BOOLEAN"
        else:
            return "VARCHAR(256)"



import dataclasses
import re
from enum import Enum

from ..direct._exceptions import ValueErrorException


class DeliveryConfigErrorMessage(Enum):
    EMAIL = """
        Invalid Email Address.
        Example:
        - email = "analyticslab.user@morningstar.com"
    """
    FTP = """
        Invalid FTP configuration.
        The username and password must not be empty
        The server must not include any protocal, like "ftp://".
        The folder must not include a root path '/'.
        Example:
        - user_name = "analyticslab_user"
        - password = "password123"
        - server = "ftp.morningstar.com"
        - folder = "folder_1/folder_2/"
    """
    DELIVERY_PROFILE = """
        Invalid delivery profile Id.
        The delivery profile id must not be empty
        The delivery profile id must be a digit string
        Example:
        - delivery_profile_id = "12345"
    """


class DeliveryType(Enum):
    EMAIL = 1
    FTP = 2
    DELIVERY_PROFILE = 3


@dataclasses.dataclass
class DeliveryConfig:
    """Container for delivery configuration settings
    Args:
        method (:obj:`DeliveryType`): Required, the type of delivery, see DeliveryType for supported options
        email_address (:obj:`str`): a valid email_address to send the file to, needed if using "email" delivery option
        user_name (:obj:`str`): the user_name to login to the ftp server, needed if using "ftp" method
        password (:obj:`str`): the password to login to the ftp server, needed if using "ftp" method
        server (:obj:`str`): the ftp server to use, needed if using "ftp" method. The server address should not include "ftp://"
        folder (:obj:`str`): the folder on the ftp server to upload the file to, needed if using "ftp" method. The folder should not include a folder path like "/folder"

    :Examples:

    Valid delivery config

    ::
        email_address = "test@morningstar.com"
        delivery_config_email = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address=email_address)

        ftp_username = "test"
        ftp_password = "test_pwd"
        ftp_server = "ts.ftp.com"
        ftp_folder = "data/"
        delivery_config_ftp = md.utils.DeliveryConfig(method=md.utils.DeliveryType.FTP, user_name=ftp_username, password=ftp_password, server=ftp_server, folder=ftp_folder)

        delivery_profile_id = "12345"
        delivery_config_email = md.utils.DeliveryConfig(method=md.utils.DeliveryType.DELIVERY_PROFILE, delivery_profile_id=delivery_profile_id)

    Invalid delivery config

    ::
        email_address = "@invalid_email_address@morningstar.com"
        delivery_config_email = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address=email_address)
        >>> ValueErrorException: Invalid Email Address

        ftp_username = "test"
        ftp_password = "test_pwd"
        ftp_server = "ftp://ts.ftp.com"
        ftp_folder = "/data/"
        delivery_config_ftp = md.utils.DeliveryConfig(method=md.utils.DeliveryType.FTP, user_name=ftp_username, password=ftp_password, server=ftp_server, folder=ftp_folder)
        >>> ValueErrorException: Invalid FTP configuration

        delivery_profile_id = "12ABC"
        delivery_config_email = md.utils.DeliveryConfig(method=md.utils.DeliveryType.DELIVERY_PROFILE, delivery_profile_id=delivery_profile_id)
        >>> ValueErrorException: Invalid delivery profile Id
    """

    method: DeliveryType
    email_address: str = ""
    delivery_profile_id: str = ""
    user_name: str = ""
    password: str = ""
    server: str = ""
    folder: str = ""

    def __post_init__(self) -> None:
        if not self.is_valid():
            raise ValueErrorException(DeliveryConfigErrorMessage[self.method.name].value)

    def _is_email_valid(self) -> bool:
        """Check whether the email config provided is valid"""
        email_regex = re.compile(
            r"""\b             # a word boundary
                [a-z0-9._%+-]+ # the email prefix part, any character in the list
                @              # the @ part
                [a-z0-9.-]+    # the email domain name part, any character in the list
                \.             # the dot part
                [a-z]{2,}      # the top-level domain part, any alphabetical character and length >= 2
                \b             # a word boundary
            """,
            re.X,
        )
        return re.fullmatch(email_regex, self.email_address.lower()) is not None

    def _is_delivery_profile_valid(self) -> bool:
        """Checks if the provided delivery profile ID is valid.
        At present, the MDS delivery endpoint allows any user to deliver to any profile ID.
        In the future, MDS need to add validation based on the user's company.
        """
        return self.delivery_profile_id.isdigit()

    def _is_ftp_valid(self) -> bool:
        """Check whether the ftp config provided is valid"""
        if self.user_name == "" or self.password == "":
            return False

        server_regex = re.compile(
            r"""\b          # a word boundary
                [a-z0-9.-]+ # the domain name part, any character in the list
                \.          # the dot part
                [a-z]{2,}   # the top-level domain part, any alphabetical character and length >= 2
                \b          # a word boundary
            """,
            re.X,
        )
        if not re.fullmatch(server_regex, self.server.lower()):
            return False

        folder_regex = re.compile(
            r"""
                ^[^\/]                   # cannot start with /
                (
                    [\w\d\s\._%+-\[\]]+  # any character in the list
                    [\/]{1,1}            # ending with /
                )+                       # capturing group
            """,
            re.X,
        )
        if not re.fullmatch(folder_regex, self.folder):
            return False

        return True

    def is_valid(self) -> bool:
        is_email_valid = self.method == DeliveryType.EMAIL and self._is_email_valid()
        is_ftp_valid = self.method == DeliveryType.FTP and self._is_ftp_valid()
        is_delivery_profile_valid = self.method == DeliveryType.DELIVERY_PROFILE and self._is_delivery_profile_valid()
        return is_email_valid or is_ftp_valid or is_delivery_profile_valid



from pandas import DataFrame
from abc import ABC, abstractmethod


class File(ABC):
    def __init__(self, df: DataFrame) -> None:
        self.df = df

    @abstractmethod
    def upload_to_s3(self, bucket: str) -> None:
        """Upload a file to an S3 bucket
        :param bucket: Bucket to upload to
        :return: True if file was uploaded, else False
        """
        pass

    @abstractmethod
    def create_s3_file_path(self) -> None:
        """
        Constructs an S3 file path comprised of a UUID and timestamp
        """
        pass



from ..._base import _logger
from ...datalake._data_objects.csvfile import CSVFile
from ...datalake._exceptions import InvalidQueryException, UnauthorizedDataLakeAccessError
from ...direct._base_api import APIBackend
from ...direct._core_api import make_request
from ...direct._error_messages import RESOURCE_NOT_FOUND_ERROR_HOLDING
from ...direct._exceptions import NetworkExceptionError, ResourceNotFoundError


import requests


import json
from typing import Any, Dict, Optional, Union


class SignedUrlBackend(APIBackend):
    """
    Subclass to call a Signed url and handle any HTTP errors that occur.
    Temporarily implements do_get_request becuase the request does not need headers and the response need not be JSON
    """

    def __init__(self) -> None:
        super().__init__()

    def do_post_request(self, url: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
        """
        Makes a POST request
        """
        return self._make_request(url=url, method="POST", data=data)

    def do_put_request(self, url: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
        """
        Makes a PUT request
        """
        return self._make_request(url=url, method="PUT", data=data)

    def do_get_request(self, url: str) -> Any:
        """
        Makes a GET request
        """
        return self._make_request(url, "GET")

    def put_csv_file(self, csv: CSVFile, signed_upload_url: str) -> Any:
        """
        Uploads a CSV file to S3 presigned url
        """
        return requests.put(url=signed_upload_url, data=csv.csv_buffer.getvalue())

    def get_bytes(self, url: str) -> Any:
        """
        Downloads bytes from S3 presigned url
        """
        res = requests.get(url)
        res.raise_for_status()
        return res

    def _make_request(self, url: str, method: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
        """
        Request and handle OK (status_code = 200) and network error responses.
        """
        _logger.info(f"Requesting: {method} {url}")
        try:
            res = make_request(method=method, url=url, headers=self.get_headers(), verify=False, data=json.dumps(data))
            res.raise_for_status()
            return res.text
        except requests.ConnectionError:
            raise NetworkExceptionError from None
        except requests.HTTPError:
            self._handle_custom_http_errors(res)
            self._handle_default_http_errors(res)

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        try:
            response_message = res.json()["message"]
        except requests.JSONDecodeError:
            response_message = res.text

        if res.status_code == 400:
            _logger.debug(f"Bad Request Error: {res.status_code} {response_message}")
            raise InvalidQueryException(response_message) from None
        if res.status_code == 401:
            _logger.debug(f"Unauthorized Error: {res.status_code} {response_message}")
            raise UnauthorizedDataLakeAccessError from None
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(RESOURCE_NOT_FOUND_ERROR_HOLDING) from None
